// Package runtime is a service runtime manager
package runtime

import (
	"errors"
	"time"
)

var (
	// DefaultRuntime is default micro runtime.
	DefaultRuntime Runtime = NewRuntime()
	// DefaultName is default runtime service name.
	DefaultName = "go.micro.runtime"

	ErrAlreadyExists = errors.New("already exists")
)

// Runtime is a service runtime manager.
type Runtime interface {
	// Init initializes runtime
	Init(...Option) error
	// Create registers a service
	Create(*Service, ...CreateOption) error
	// Read returns the service
	Read(...ReadOption) ([]*Service, error)
	// Update the service in place
	Update(*Service, ...UpdateOption) error
	// Remove a service
	Delete(*Service, ...DeleteOption) error
	// Logs returns the logs for a service
	Logs(*Service, ...LogsOption) (LogStream, error)
	// Start starts the runtime
	Start() error
	// Stop shuts down the runtime
	Stop() error
	// String describes runtime
	String() string
}

// Stream returns a log stream.
type LogStream interface {
	Error() error
	Chan() chan LogRecord
	Stop() error
}

type LogRecord struct {
	Metadata map[string]string
	Message  string
}

// Scheduler is a runtime service scheduler.
type Scheduler interface {
	// Notify publishes schedule events
	Notify() (<-chan Event, error)
	// Close stops the scheduler
	Close() error
}

// EventType defines schedule event.
type EventType int

const (
	// Create is emitted when a new build has been craeted.
	Create EventType = iota
	// Update is emitted when a new update become available.
	Update
	// Delete is emitted when a build has been deleted.
	Delete
)

// String returns human readable event type.
func (t EventType) String() string {
	switch t {
	case Create:
		return "create"
	case Delete:
		return "delete"
	case Update:
		return "update"
	default:
		return "unknown"
	}
}

// Event is notification event.
type Event struct {
	// Timestamp is event timestamp
	Timestamp time.Time
	// Service the event relates to
	Service *Service
	// Options to use when processing the event
	Options *CreateOptions
	// ID of the event
	ID string
	// Type is event type
	Type EventType
}

// Service is runtime service.
type Service struct {
	// Metadata stores metadata
	Metadata map[string]string
	// Name of the service
	Name string
	// Version of the service
	Version string
	// url location of source
	Source string
}
